#ifndef _DELAUNAY_TET_MESH_H_
#define _DELAUNAY_TET_MESH_H_

#include "MyTMesh.h"

namespace DartLib
{
namespace PowerDiagram3D
{
// Customize CPDTet_3 for class CDelaunayTMesh
class CDelaunayTet_3 : public CMyTet_3
{
    // the dual of the tetrahedron
    ADD_TRAIT(CPoint, dual)
    ADD_TRAIT(double, radius)
};

template <typename V, typename E, typename F, typename T, typename Dart>
class TDelaunayTMesh : public TMyTMesh<V, E, F, T, Dart>
{
  public:
    /*! 
     *  Simply use std::vector as the container for dual face and cell.
     */
    using CDualFace = std::vector<CPoint>;
    using CDualCell = std::vector<CDualFace>;

    /*!
     *  Compute the dual cell for a vertex.
     *  \param [in] pV: the input vertex.
     *  \return the dual cell which consists of a list of dual faces.
     */
    CDualCell vertex_dual_cell(V* pV);

    /*!
     *  Compute the dual face for an edge.
     *  \param [in] pE: the input edge.
     *  \return the dual face which consists of a list of dual vertices.
     */
    CDualFace edge_dual_face(E* pE);
    
    /*!
     *  Compute the normal for a dual face.
     *  \param [in] dual_face: the input dual face.
     *  \return the normal
     */
    CPoint dual_face_normal(const CDualFace& dual_face);
};


using CDelaunayTMesh = TDelaunayTMesh<CMyVertex_3, CEdge_3, CMyFace_3, CDelaunayTet_3, CDart_3>;

template <typename V, typename E, typename F, typename T, typename Dart>
CDelaunayTMesh::CDualCell TDelaunayTMesh<V, E, F, T, Dart>::vertex_dual_cell(V* pV)
{
    CDualCell dual_cell;

    //insert your coder here
    //use CDelaunayTMesh::VertexEdgeIterator
    //each edge corresponding to a dual face

    return dual_cell;
}

template <typename V, typename E, typename F, typename T, typename Dart>
CDelaunayTMesh::CDualFace TDelaunayTMesh<V, E, F, T, Dart>::edge_dual_face(E* pE)
{
    CDualFace dual_face;

    //insert your coder here
    //use CDelaunayTMesh::EdgeVolumeIterator
    //for boundary edge, specially handle them using face normal, compute_face_normal(pFace)

    return dual_face;
}

template <typename V, typename E, typename F, typename T, typename Dart>
CPoint TDelaunayTMesh<V, E, F, T, Dart>::dual_face_normal(const CDualFace& dual_face)
{
    CPoint pts[3];

    int i = 0;
    for (const auto& p : dual_face)
    {
        pts[i++] = p;
        if (i >= 3)
            break;
    }
    CPoint n = (pts[1] - pts[0]) ^ (pts[2] - pts[0]);
    n /= n.norm();

    return n;
}

} // namespace PowerDiagram3D
} // namespace DartLib

#endif //! _DELAUNAY_TET_MESH_H_
